#include<stdio.h>
#include<stdlib.h>
#include<math.h>
#include"SimpleImage.h" /* 画像クラス */
#include"BMPIO.h" /* BMP画像の入出力関数 stdlib.hが必要 */

/* 正規化用の関数 */
void Normalize(Image *R,Image *G,Image *B){
  double max,min;
  max= min = R->img[0][0];
  int i,j;
  
  for(i=0;i<R->sy;i++)
    for(j=0;j<R->sx;j++){
      if(max<R->img[i][j])max=R->img[i][j];
      if(max<G->img[i][j])max=G->img[i][j];
      if(max<B->img[i][j])max=B->img[i][j];
      if(min>R->img[i][j])min=R->img[i][j];
      if(min>G->img[i][j])min=G->img[i][j];
      if(min>B->img[i][j])min=B->img[i][j];
    }
  double range = fabs(max-min);if(range<=0.0)range=1.0;
   
  for(i=0;i<R->sy;i++)
     for(j=0;j<R->sx;j++){
       double val = 255.0*(R->img[i][j]-min)/range;if(val<0.0)val=0.0;if(val>255.0)val=255.0;
      R->img[i][j] = val;
      val = 255.0*(G->img[i][j]-min)/range;if(val<0.0)val=0.0;if(val>255.0)val=255.0;
      G->img[i][j] = val;
      val = 255.0*(B->img[i][j]-min)/range;if(val<0.0)val=0.0;if(val>255.0)val=255.0;
      B->img[i][j] = val;
     }
     
}

/* BMP画像を読み込んでGaussianフィルタをセパレート実装で適用しBMP画像として保存する 
コンパイル: g++ -o SeparableGaussian SeparableGaussian.cxx -lm
実行： ./SeparableGaussian 入力BMP 出力BMP Gauss関数のsigma 畳み込み半径(int)
引数４
*/
int main(int argc,char *argv[]){
  
  
  /* 画像クラスImageを用いて入力画像のカラー用にR,G,Bを宣言する */
  Image *R = NULL;
  Image *G = NULL;
  Image *B = NULL;
  

  /* 画像サイズ */
  int sx,sy;

  
  /* BMPIOクラスを宣言しnewする */
  BMPIO *mybmp = new BMPIO();
  /* BMPIOクラスのreadBMPSize()を使って、argv[1]で与えられたファイル名でsx,syに画像サイズを入れる */
  if(mybmp->readBMPSize(&sx,&sy,argv[1])){

    /* sx,syでR,G,Bのメモリを確保する(new する) */
    R = new Image(sx,sy);
    G = new Image(sx,sy);
    B = new Image(sx,sy);

    /* BMPIOクラスのreadBMP()を使ってargv[1]で与えられるファイル名の画像をR,G,Bに読み込む */
    mybmp->readBMP(R,G,B,argv[1]);
    
  }else{
    delete mybmp;
    printf("Unsupported BMP format !\n");
    return 1;
  }
  
  
 
  /* Gauss関数のパラメータをsigmaへセットする */
  double sigma = atof(argv[3]);
  double sigI = 1.0/(2.0*sigma*sigma);
  /* 畳み込み半径をradiiへセットする */
  int radii = atoi(argv[4]);
    

  /* 出力画像用にImageクラスをoutR,outG,outBとして宣言しsx,syのサイズでメモリを確保する */
  Image *outR = new Image(sx,sy);
  Image *outG = new Image(sx,sy);
  Image *outB = new Image(sx,sy);
    
  /* セパレートガウスフィルタ計算ここから */
   int i,j;
  /* 一次元配列でガウス画像を用意する　*/
  int ksize = 2*radii+1;
  double *kernel = new double[ksize];
  double sum=0.0;
  for(i=-radii;i<=radii;i++){
    double weight = exp(-((double)(i*i))*sigI);
    kernel[i+radii] = weight;
    sum += weight;
  }
  if(sum<=0.0)sum=1.0;
  for(i=0;i<ksize;i++)kernel[i]/=sum;
  /* テンポラリーの画像を用意する */  
  Image *tmpR = new Image(sx,sy);
  Image *tmpG = new Image(sx,sy);
  Image *tmpB = new Image(sx,sy);
  


  /* X方向のフィルタ */
  for(i=0;i<sy;i++)
      for(j=0;j<sx;j++){
	tmpR->img[i][j] =0.0;
	tmpG->img[i][j] =0.0;
	tmpB->img[i][j] =0.0;
	for(int k=-radii;k<=radii;k++){
	  int jk = k+j;
	  if(jk<0)jk=0;
	  if(jk>=sx)jk=sx-1;
	  double weight = kernel[k+radii];
	  tmpR->img[i][j] +=R->img[i][jk]*weight;
	  tmpG->img[i][j] +=G->img[i][jk]*weight;
	  tmpB->img[i][j] +=B->img[i][jk]*weight;
	}
      }
  /* Y方向のフィルタ */
  for(j=0;j<sx;j++)
      for(i=0;i<sy;i++){
	outR->img[i][j] =0.0;
	outG->img[i][j] =0.0;
	outB->img[i][j] =0.0;
	for(int k=-radii;k<=radii;k++){
	  int ik = k+i;
	  if(ik<0)ik=0;
	  if(ik>=sy)ik=sy-1;
	  double weight = kernel[k+radii];
	  outR->img[i][j] +=tmpR->img[ik][j]*weight;
	  outG->img[i][j] +=tmpG->img[ik][j]*weight;
	  outB->img[i][j] +=tmpB->img[ik][j]*weight;
	}
      }
    
  /* Gaussian計算ここまで */
  /* outR,outG,outBを0-255へ線形補間で正規化する */
  Normalize(outR,outG,outB);
  
  
  /* BMPIOクラスのsaveBMP()を使ってargv[2]で与えられるファイル名にoutR,outG,outBの画像を保存する */
  mybmp->saveBMP(outR,outG,outB,argv[2]);
  
  
  /* BMPIOクラスのメモリを開放する(deleteする) */
  delete mybmp;
  
  /* R,G,Bのメモリを開放する(deleteする) */
  delete R;
  delete G;
  delete B;
  /* テンポラリーR,G,Bのメモリを開放する(deleteする) */
  delete tmpR;
  delete tmpG;
  delete tmpB;
  
  /* outR,outG,outBのメモリを開放する(deleteする) */
  delete outR;
  delete outG;
  delete outB;

  /*  ガウス画像 のメモリを開放する(deleteする) */
  delete [] kernel;
 


  return 0;
}
